/* Copyright (C) |Meso|Star> 2016 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "scpr.h"
#include "test_scpr_utils.h"

#include <rsys/math.h>
#include <rsys/stretchy_array.h>

static void
get_clip_pos(const size_t ivert, double pos[2], void* ctx)
{
  const double* coords = ctx;
  NCHECK(pos, NULL);
  NCHECK(ctx, NULL);
  pos[0] = coords[ivert*2+0];
  pos[1] = coords[ivert*2+1];
}

static void
dump_obj(FILE* stream, const struct scpr_mesh* mesh)
{
  size_t i, n;

  NCHECK(stream, NULL);
  NCHECK(mesh, NULL);

  CHECK(scpr_mesh_get_vertices_count(mesh, &n), RES_OK);
  FOR_EACH(i, 0, n) {
    double pos[2];
    CHECK(scpr_mesh_get_position(mesh, i, pos), RES_OK);
    fprintf(stream, "v %g %g 0\n", SPLIT2(pos));
  }

  CHECK(scpr_mesh_get_triangles_count(mesh, &n), RES_OK);
  FOR_EACH(i, 0, n) {
    size_t ids[3];
    CHECK(scpr_mesh_get_indices(mesh, i, ids), RES_OK);
    fprintf(stream, "f %lu %lu %lu\n",
      (unsigned long)(ids[0] + 1),
      (unsigned long)(ids[1] + 1),
      (unsigned long)(ids[2] + 1));
  }
}

static void
test_triangle(struct scpr_mesh* mesh)
{
  const double triangle_pos[] = { 0.0, 0.0, 0.0, 1.0, 1.0, 0.0 };
  const size_t triangle_ids[] = { 0, 1, 2 };
  const double clip_pos[] = { -1.0, 0.25, 1.0, 0.75, 1, 0.25 };
  struct scpr_polygon poly;
  struct mesh_context ctx;
  size_t ntris;

  ctx.coords = triangle_pos;
  ctx.nverts = 3;
  ctx.indices = triangle_ids;
  ctx.ntris = 1;
  CHECK(scpr_mesh_setup_indexed_vertices
    (mesh, ctx.ntris, get_ids, 3, get_pos, &ctx), RES_OK);

  poly.get_position = get_clip_pos;
  poly.nvertices = sizeof(clip_pos)/sizeof(double[2]);
  poly.context = (void*)clip_pos;
  CHECK(scpr_mesh_clip(NULL, SCPR_OPERATIONS_COUNT__, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_clip(mesh, SCPR_OPERATIONS_COUNT__, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_clip(NULL, SCPR_OPERATIONS_COUNT__, &poly), RES_BAD_ARG);
  CHECK(scpr_mesh_clip(mesh, SCPR_OPERATIONS_COUNT__, &poly), RES_BAD_ARG);
  CHECK(scpr_mesh_clip(NULL, SCPR_SUB, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_clip(mesh, SCPR_SUB, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_clip(NULL, SCPR_SUB, &poly), RES_BAD_ARG);
  CHECK(scpr_mesh_clip(mesh, SCPR_SUB, &poly), RES_OK);

  /*dump_obj(stdout, mesh);*/

  CHECK(scpr_mesh_get_triangles_count(mesh, &ntris), RES_OK);
  CHECK(ntris, 3);
}

static void
test_quad(struct scpr_mesh* mesh)
{
  const double quad_pos[] = { 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0 };
  const size_t quad_ids[] = { 0, 1, 3, 3, 1, 2 };
  const double clip_pos[] = { -0.25, 0.25, -0.25, 0.75, 1.25, 0.75, 1.25, 0.25 };
  struct scpr_polygon poly;
  struct mesh_context ctx;

  ctx.coords = quad_pos;
  ctx.nverts = sizeof(quad_pos)/sizeof(double[2]);
  ctx.indices = quad_ids;
  ctx.ntris = sizeof(quad_ids)/sizeof(size_t[3]);
  CHECK(scpr_mesh_setup_indexed_vertices
    (mesh, ctx.ntris, get_ids, ctx.nverts, get_pos, &ctx), RES_OK);

  poly.get_position = get_clip_pos;
  poly.nvertices = sizeof(clip_pos)/sizeof(double[2]);
  poly.context = (void*)clip_pos;
  CHECK(scpr_mesh_clip(mesh, SCPR_AND, &poly), RES_OK);

  /*dump_obj(stdout, mesh);*/
}

static void
test_disk(struct scpr_mesh* mesh)
{
  const double clip[] =  { -1.75, -1.75, 1.75, -1.75, 1.75, 1.75, -1.75, 1.75 };
  const size_t ninternal_disks = 10;
  const double radius = 2.5;
  const double internal_disk_step = radius / (double)ninternal_disks;
  const size_t nslices = 64;
  struct mesh_context ctx;
  struct scpr_polygon poly;
  double* pos = NULL;
  size_t* ids = NULL;
  size_t i, j;

  FOR_EACH(i, 0, ninternal_disks) {
    const double r = (double)(i+1)*internal_disk_step;
    FOR_EACH(j, 0, nslices) {
      const double theta = (double)j / (double)nslices * 2 * PI;
      const double x = r * cos(theta);
      const double y = r * sin(theta);
      sa_push(pos, x);
      sa_push(pos, y);
    }
  }

  /* Center point */
  sa_push(pos, 0.0);
  sa_push(pos, 0.0);

  FOR_EACH(i, 0, ninternal_disks-1) {
    const size_t offset = (i+1) * nslices;
    FOR_EACH(j, 0, nslices) {
      const size_t id0 = j + offset;
      const size_t id1 = ((j + 1) % nslices) + offset;
      const size_t id2 = id0 - nslices;
      const size_t id3 = id1 - nslices;

      sa_push(ids, id0);
      sa_push(ids, id2);
      sa_push(ids, id1);

      sa_push(ids, id1);
      sa_push(ids, id2);
      sa_push(ids, id3);
    }
  }

  /* Center triangles */
  FOR_EACH(j, 0, nslices) {
    const size_t id0 = j;
    const size_t id1 = (j + 1) % nslices;
    const size_t id2 = sa_size(pos)/2 - 1;

    sa_push(ids, id0);
    sa_push(ids, id2);
    sa_push(ids, id1);
  }

  ctx.coords = pos;
  ctx.nverts = sa_size(pos)/2;
  ctx.indices = ids;
  ctx.ntris = sa_size(ids)/3;
  CHECK(scpr_mesh_setup_indexed_vertices
    (mesh, ctx.ntris, get_ids, ctx.nverts, get_pos, &ctx), RES_OK);

  poly.get_position = get_clip_pos;
  poly.nvertices = sizeof(clip)/sizeof(double[2]);
  poly.context = (void*)clip;
  CHECK(scpr_mesh_clip(mesh, SCPR_SUB, &poly), RES_OK);

  dump_obj(stdout, mesh);

  sa_release(pos);
  sa_release(ids);
}

int
main(int argc, char** argv)
{
  struct mem_allocator allocator;
  struct scpr_mesh* mesh;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHECK(scpr_mesh_create(&allocator, &mesh), RES_OK);

  test_triangle(mesh);
  test_quad(mesh);
  test_disk(mesh);

  CHECK(scpr_mesh_ref_put(mesh), RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHECK(mem_allocated_size(), 0);
  return 0;
}

