/* Copyright (C) |Meso|Star> 2016-2017 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#define _POSIX_C_SOURCE 200112L /* nextafter */

#include "ssf.h"
#include "ssf_bxdf_c.h"

#include <rsys/double3.h>
#include <rsys/double4.h>
#include <rsys/mem_allocator.h>
#include <rsys/ref_count.h>

#include <star/ssp.h>

#define MAX_BxDFs 8

struct ssf_bsdf {
  struct ssf_bxdf* bxdfs[MAX_BxDFs];

  double uweights[MAX_BxDFs]; /* Unormalized component weights */
  double weights[MAX_BxDFs]; /* Normalized weights */
  int outdated_weights; /* Define if the normalized weights are outdated */

  size_t nbxdfs;

  ref_T ref;
  struct mem_allocator* allocator;
};

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static INLINE void
bsdf_normalize_weights(struct ssf_bsdf* bsdf)
{
  double weights_sum = 0;
  size_t i;
  ASSERT(bsdf);
  if(!bsdf->outdated_weights) return;
  FOR_EACH(i, 0, bsdf->nbxdfs) weights_sum += bsdf->uweights[i];
  FOR_EACH(i, 0, bsdf->nbxdfs) bsdf->weights[i] = bsdf->uweights[i]/weights_sum;
  bsdf->outdated_weights = 0;
}

static void
bsdf_release(ref_T* ref)
{
  struct ssf_bsdf* bsdf = CONTAINER_OF(ref, struct ssf_bsdf, ref);
  ASSERT(ref);
  SSF(bsdf_clear(bsdf));
  MEM_RM(bsdf->allocator, bsdf);
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
ssf_bsdf_create(struct mem_allocator* allocator, struct ssf_bsdf** out_bsdf)
{
  struct mem_allocator* mem_allocator = NULL;
  struct ssf_bsdf* bsdf = NULL;
  res_T res = RES_OK;

  if(!out_bsdf) {
    res = RES_BAD_ARG;
    goto error;
  }

  mem_allocator = allocator ? allocator : &mem_default_allocator;
  bsdf = MEM_CALLOC(mem_allocator, 1, sizeof(struct ssf_bsdf));
  if(!bsdf) {
    res = RES_MEM_ERR;
    goto error;
  }
  bsdf->allocator = mem_allocator;
  ref_init(&bsdf->ref);

exit:
  if(out_bsdf) *out_bsdf = bsdf;
  return res;
error:
  if(bsdf) {
    SSF(bsdf_ref_put(bsdf));
    bsdf = NULL;
  }
  goto exit;
}

res_T
ssf_bsdf_ref_get(struct ssf_bsdf* bsdf)
{
  if(!bsdf) return RES_BAD_ARG;
  ref_get(&bsdf->ref);
  return RES_OK;
}

res_T
ssf_bsdf_ref_put(struct ssf_bsdf* bsdf)
{
  if(!bsdf) return RES_BAD_ARG;
  ref_put(&bsdf->ref, bsdf_release);
  return RES_OK;
}

res_T
ssf_bsdf_add(struct ssf_bsdf* bsdf, struct ssf_bxdf* bxdf, const double weight)
{
  if(!bsdf || !bxdf || weight < 0) return RES_BAD_ARG;
  if(bsdf->nbxdfs >= MAX_BxDFs) return RES_MEM_ERR;
  if(weight == 0) return RES_OK;
  SSF(bxdf_ref_get(bxdf));
  bsdf->bxdfs[bsdf->nbxdfs] = bxdf;
  bsdf->uweights[bsdf->nbxdfs] = weight;
  bsdf->outdated_weights = 1;
  ++bsdf->nbxdfs;
  return RES_OK;
}

res_T
ssf_bsdf_clear(struct ssf_bsdf* bsdf)
{
  size_t i;
  if(!bsdf) return RES_BAD_ARG;
  FOR_EACH(i, 0, bsdf->nbxdfs) {
    SSF(bxdf_ref_put(bsdf->bxdfs[i]));
  }
  bsdf->nbxdfs = 0;
  return RES_OK;
}

double
ssf_bsdf_sample
  (struct ssf_bsdf* bsdf,
   struct ssp_rng* rng,
   const double wo[3],
   const double N[3],
   double wi[3],
   int* type,
   double* out_pdf)
{
  double cumul;
  double R;
  double r;
  double pdf;
  size_t i;
  ASSERT(bsdf && rng && wi && N && wo && type && out_pdf);
  ASSERT(d3_is_normalized(wo) && d3_is_normalized(N));

  if(!bsdf->nbxdfs) {
    *out_pdf = 0;
    *type = 0;
    return 0;
  }
  if(bsdf->nbxdfs == 1) {
    return ssf_bxdf_sample(bsdf->bxdfs[0], rng, wo, N, wi, type, out_pdf);
  }

  bsdf_normalize_weights(bsdf);

  /* Sample a component */
  r = ssp_rng_canonical(rng);
  cumul = bsdf->weights[0];
  FOR_EACH(i, 0, bsdf->nbxdfs-1) {
    if(r <= cumul) break;
    cumul += bsdf->weights[i];
  }

  /* Sample a direction from the selected component */
  R = ssf_bxdf_sample(bsdf->bxdfs[i], rng, wo, N, wi, type, &pdf);
  *out_pdf = pdf*bsdf->weights[i];
  return R;
}

double
ssf_bsdf_pdf
  (struct ssf_bsdf* bsdf,
   const double wo[3],
   const double N[3],
   const double wi[3])
{
  double pdf = 0;
  size_t i;
  ASSERT(bsdf);
  bsdf_normalize_weights(bsdf);
  FOR_EACH(i, 0, bsdf->nbxdfs) {
    pdf += ssf_bxdf_pdf(bsdf->bxdfs[i], wo, N, wi) * bsdf->weights[i];
  }
  return pdf;
}

double
ssf_bsdf_eval
  (struct ssf_bsdf* bsdf,
   const double wo[3],
   const double N[3],
   const double wi[3])
{
  double fr = 0;
  size_t i;
  ASSERT(bsdf);
  bsdf_normalize_weights(bsdf);
  FOR_EACH(i, 0, bsdf->nbxdfs) {
    fr += bsdf->weights[i] * ssf_bxdf_eval(bsdf->bxdfs[i], wo, N, wi);
  }
  return fr;
}

