/* Copyright (C) |Meso|Star> 2016-2017 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "ssf.h"
#include "test_ssf_utils.h"

#include <rsys/double3.h>

static int bxdf_is_init = 0;

struct ALIGN(64) bxdf {
  uint32_t id;
  struct ssp_rng* rng;
  double wi[3];
  double wo[3];
  double N[3];
  double reflectivity;
  double value;
  double pdf;
  double rho;
};

static res_T
bxdf_init(struct mem_allocator* allocator, void* bxdf)
{
  NCHECK(allocator, NULL);
  NCHECK(bxdf, NULL);
  CHECK(IS_ALIGNED(bxdf, 64), 1);
  ((struct bxdf*)bxdf)->id = 0xDECAFBAD;
  bxdf_is_init = 1;
  return RES_OK;
}

static void
bxdf_release(void* bxdf)
{
  NCHECK(bxdf, NULL);
  CHECK(((struct bxdf*)bxdf)->id, 0xDECAFBAD);
  bxdf_is_init = 0;
}

static double
bxdf_sample
  (void* bxdf,
   struct ssp_rng* rng,
   const double wo[3],
   const double N[3],
   double wi[3],
   int* type,
   double* pdf)
{
  struct bxdf* BxDF = bxdf;
  NCHECK(BxDF, NULL);
  CHECK(BxDF->rng, rng);
  CHECK(d3_eq(BxDF->wo, wo), 1);
  CHECK(d3_eq(BxDF->N, N), 1);
  d3(wi, 1.0, 2.0, 3.0);
  *type = 314;
  *pdf = 4;
  return BxDF->reflectivity;
}

static double
bxdf_eval
  (void* bxdf,
   const double wo[3],
   const double N[3],
   const double wi[3])
{
  struct bxdf* BxDF = bxdf;
  NCHECK(BxDF, NULL);
  NCHECK(wi, NULL);
  NCHECK(wo, NULL);
  CHECK(d3_eq(BxDF->wo, wo), 1);
  CHECK(d3_eq(BxDF->N, N), 1);
  CHECK(d3_eq(BxDF->wi, wi), 1);
  return BxDF->value;
}

static double
bxdf_pdf
  (void* bxdf,
   const double wo[3],
   const double N[3],
   const double wi[3])
{
  struct bxdf* BxDF = bxdf;
  NCHECK(BxDF, NULL);
  NCHECK(wi, NULL);
  NCHECK(wo, NULL);
  CHECK(d3_eq(BxDF->wo, wo), 1);
  CHECK(d3_eq(BxDF->N, N), 1);
  CHECK(d3_eq(BxDF->wi, wi), 1);
  return BxDF->pdf;
}

int
main(int argc, char** argv)
{
  struct mem_allocator allocator;
  struct ssp_rng* rng;
  struct bxdf* data;
  struct ssf_bxdf* bxdf;
  struct ssf_bxdf_type type = SSF_BXDF_TYPE_NULL;
  double wo[3];
  double N[3];
  double wi[4];
  double pdf;
  int i;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);
  CHECK(ssp_rng_create(&allocator, &ssp_rng_threefry, &rng), RES_OK);

  type.init = bxdf_init;
  type.release = bxdf_release;
  type.sample = bxdf_sample;
  type.eval = bxdf_eval;
  type.pdf = bxdf_pdf;
  type.sizeof_bxdf = sizeof(struct bxdf);
  type.alignof_bxdf = ALIGNOF(struct bxdf);

  CHECK(ssf_bxdf_create(NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(ssf_bxdf_create(&allocator, NULL, NULL), RES_BAD_ARG);
  CHECK(ssf_bxdf_create(NULL, &type, NULL), RES_BAD_ARG);
  CHECK(ssf_bxdf_create(&allocator, &type, NULL), RES_BAD_ARG);
  CHECK(ssf_bxdf_create(NULL, NULL, &bxdf), RES_BAD_ARG);
  CHECK(ssf_bxdf_create(&allocator, NULL, &bxdf), RES_BAD_ARG);

  CHECK(bxdf_is_init, 0);
  CHECK(ssf_bxdf_create(NULL, &type, &bxdf), RES_OK);
  CHECK(bxdf_is_init, 1);

  CHECK(ssf_bxdf_ref_get(NULL), RES_BAD_ARG);
  CHECK(ssf_bxdf_ref_get(bxdf), RES_OK);
  CHECK(ssf_bxdf_ref_put(NULL), RES_BAD_ARG);
  CHECK(ssf_bxdf_ref_put(bxdf), RES_OK);
  CHECK(bxdf_is_init, 1);
  CHECK(ssf_bxdf_ref_put(bxdf), RES_OK);
  CHECK(bxdf_is_init, 0);

  CHECK(ssf_bxdf_create(&allocator, &type, &bxdf), RES_OK);
  CHECK(bxdf_is_init, 1);
  CHECK(ssf_bxdf_ref_put(bxdf), RES_OK);
  CHECK(bxdf_is_init, 0);

  type.init = NULL;
  CHECK(ssf_bxdf_create(&allocator, &type, &bxdf), RES_BAD_ARG);
  CHECK(bxdf_is_init, 0);
  type.init = bxdf_init;
  type.release = NULL;
  CHECK(ssf_bxdf_create(&allocator, &type, &bxdf), RES_BAD_ARG);
  CHECK(bxdf_is_init, 0);
  type.release = bxdf_release;
  type.sample = NULL;
  CHECK(ssf_bxdf_create(&allocator, &type, &bxdf), RES_BAD_ARG);
  CHECK(bxdf_is_init, 0);
  type.sample = bxdf_sample;
  type.alignof_bxdf = 3;
  CHECK(ssf_bxdf_create(&allocator, &type, &bxdf), RES_BAD_ARG);
  CHECK(bxdf_is_init, 0);
  type.alignof_bxdf = ALIGNOF(struct bxdf);
  CHECK(ssf_bxdf_create(&allocator, &type, &bxdf), RES_OK);
  CHECK(bxdf_is_init, 1);

  CHECK(ssf_bxdf_get_data(NULL, NULL), RES_BAD_ARG);
  CHECK(ssf_bxdf_get_data(bxdf, NULL), RES_BAD_ARG);
  CHECK(ssf_bxdf_get_data(NULL, (void**)&data), RES_BAD_ARG);
  CHECK(ssf_bxdf_get_data(bxdf, (void**)&data), RES_OK);

  CHECK(data->id, 0xDECAFBAD);

  d3_normalize(wo, d3(wo, -1, -1, 0));
  d3(N, 0.0, 1.0, 0.0);
  d3_set(data->wo, wo);
  d3_set(data->N, N);
  data->rng = rng;
  data->reflectivity = 0.1234;

  CHECK(ssf_bxdf_sample(bxdf, rng, wo, N, wi, &i, &pdf), 0.1234);
  CHECK(ssf_bxdf_sample(bxdf, rng, wo, N, wi, &i, &pdf), 0.1234);
  CHECK(i, 314);
  CHECK(pdf, 4);

  data->reflectivity = 0.314;
  CHECK(ssf_bxdf_sample(bxdf, rng, wo, N, wi, &i, &pdf), 0.314);

  d3_normalize(wi, wi);
  d3_set(data->wi, wi);
  data->value = 0.4567;
  CHECK(ssf_bxdf_eval(bxdf, wo, N, wi), data->value);
  data->pdf = 0.890;
  CHECK(ssf_bxdf_pdf(bxdf, wo, N, wi), data->pdf);

  CHECK(bxdf_is_init, 1);
  CHECK(ssf_bxdf_ref_put(bxdf), RES_OK);
  CHECK(bxdf_is_init, 0);

  CHECK(ssp_rng_ref_put(rng), RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHECK(mem_allocated_size(), 0);
  return 0;
}

